# -*- coding: utf-8 -*-
"""
Summary
-------

Functions to compare and plot comparison of :class:`~radis.spectrum.spectrum.Spectrum` 
objects


Routine Listings
----------------

- :func:`~radis.spectrum.compare.get_diff`
- :func:`~radis.spectrum.compare.get_distance`
- :func:`~radis.spectrum.compare.get_ratio`
- :func:`~radis.spectrum.compare.get_residual`
- :func:`~radis.spectrum.compare.get_residual_integral`
- :func:`~radis.spectrum.compare.plot_diff`


-------------------------------------------------------------------------------



"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.misc.arrays import array_allclose
from radis.misc.curve import curve_substract, curve_distance, curve_divide
from radis.spectrum.spectrum import make_up, cast_waveunit
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.widgets import MultiCursor
import numpy as np
from publib import set_style, fix_style

# %% ======================================================================
# External functions
# ----------------
# XXX =====================================================================


def get_diff(s1, s2, var, wunit='default', Iunit='default', medium='default',
             resample=True):
    ''' Get the difference between 2 spectra
    Basically returns w1, I1 - I2 where (w1, I1) and (w2, I2) are the values of
    s1 and s2 for variable var. (w2, I2) is linearly interpolated if needed.


    Parameters    
    ----------

    s1, s2: Spectrum objects

    var: str
        spectral quantity (ex: 'radiance', 'transmittance'...)

    wunit: 'nm', 'cm-1'
        waveunit to compare in

    Iunit: str
        if ``'default'`` use s1 unit for variable var

    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)

    Returns    
    -------

    w1, Idiff: array
        difference interpolated on the second range(order?)
    
    Notes
    -----
    
    Uses :func:`~radis.misc.curve.curve_substract` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :func:`~radis.spectrum.compare.plot_diff` 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''

    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit,
                                   medium=medium, assert_same_wavelength=not resample)

    # basically w1, I1 - I2 (on same range)
    return curve_substract(w1, I1, w2, I2)


def get_ratio(s1, s2, var, wunit='default', Iunit='default', medium='default',
              resample=True):
    ''' Get the ratio between two spectra
    Basically returns w1, I1 / I2 where (w1, I1) and (w2, I2) are the values of
    s1 and s2 for variable var. (w2, I2) is linearly interpolated if needed.


    Parameters    
    ----------

    s1, s2: Spectrum objects

    var: str
        spectral quantity 

    wunit: 'nm', 'cm-1'
        waveunit to compare in

    Iunit: str
        if ``'default'`` use s1 unit for variable var

    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)

    Notes
    -----
    
    Uses :func:`~radis.misc.curve.curve_divide` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_distance`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with` 
    

    '''

    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit,
                                   medium=medium, assert_same_wavelength=not resample)

    # basically w1, I1 - I2 (on same range)
    return curve_divide(w1, I1, w2, I2)


def get_distance(s1, s2, var, wunit='default', Iunit='default', medium='default',
                 resample=True):
    ''' Get the Euclidian distance between 2 spectra

    If waveranges dont match, `s2` is interpolated over `s1`. 


    Parameters    
    ----------

    s1, s2: Spectrum objects

    var: str
        spectral quantity 

    wunit: 'nm', 'cm-1'
        waveunit to compare in

    Iunit: str
        if ``'default'`` use s1 unit for variable var

    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)
    
    Notes
    -----
    
    Uses :func:`~radis.misc.curve.curve_distance` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with` 

    '''
    # TODO: normalize with Imax, wmax 

    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit,
                                   medium=medium, assert_same_wavelength=not resample)

    # euclidian distance from w1, I1
    return curve_distance(w1, I1, w2, I2, discard_out_of_bounds=True)


def get_residual(s1, s2, var, norm='L2', ignore_nan=False):
    ''' Returns L2 norm of ``s1`` and ``s2``

    For ``I1``, ``I2``, the values of variable ``var`` in ``s1`` and ``s2``, 
    respectively, residual is calculated as:
        
    For ``L2`` norm::

        np.sqrt((dI**2).sum())/len(dI)
        
    For ``L1`` norm::
        
        (np.abs(dI)).sum()/len(dI)


    Parameters    
    ----------

    s1, s2: :class:`~radis.spectrum.spectrum.Spectrum` objects
        if not on the same range, ``s2`` is resampled on ``s1``.

    var: str
        spectral quantity

    norm: 'L2', 'L1'
        which norm to use 

    Other Parameters
    ----------------

    ignore_nan: boolean
        if ``True``, ignore nan in the difference between s1 and s2 (ex: out of bound)
        when calculating residual. Default ``False``. Note: ``get_residual`` will still 
        fail if there are nan in initial Spectrum. 

    Notes
    -----

    0 values for I1 yield nans except if I2 = I1 = 0

    when s1 and s2 dont have the size wavespace range, they are automatically
    resampled through get_diff on 's1' range 


    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.plot_diff`, 
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''

    w, I = s1.get(var)
    # mask for 0
    wdiff, dI = get_diff(s1, s2, var, resample=True)

    if ignore_nan:
        b = np.isnan(dI)
        wdiff, dI = wdiff[~b], dI[~b]

    if norm == 'L2':
        return np.sqrt((dI**2).sum())/len(dI)
    elif norm == 'L1':
        return (np.abs(dI)).sum()/len(dI)
    else:
        raise ValueError('unexpected value for norm')


def get_residual_integral(s1, s2, var, ignore_nan=False):
    ''' Returns integral of the difference between two spectra s1 and s2, 
    relatively to the integral of spectrum s1 
    
    Compared to :func:`~radis.spectrum.compare.get_residual`, this tends to 
    cancel the effect of the gaussian noise of an experimental spectrum. 
    
    ::
    
        res = trapz(I2-I1, w1) / trapz(I1, w1)

    Note: when the considered variable is ``transmittance`` or ``transmittance_noslit``, 
    the *upper* integral is used (up to 1) to normalize the integral difference
    
    ::
    
        res = trapz(I2-I1, w1) / trapz(1-I1, w1)

    Parameters
    ----------

    s1, s2: Spectrum objects

    var: str
        spectral quantity

    Other Parameters
    ----------------

    ignore_nan: boolean
        if ``True``, ignore nan in the difference between s1 and s2 (ex: out of bound)
        when calculating residual. Default ``False``. Note: ``get_residual_integral``
        will still fail if there are nan in initial Spectrum. 

    Notes
    -----

    For I1, I2, the values of 'var' in s1 and s2, respectively, residual
    is calculated as::

        res = trapz(I2-I1, w1) / trapz(I1, w1)

    0 values for I1 yield nans except if I2 = I1 = 0

    when s1 and s2 dont have the size wavespace range, they are automatically
    resampled through get_diff on 's1' range 


    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.get_residual`, 
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''

    w, I = s1.get(var)
    # mask for 0
    wdiff, dI = get_diff(s1, s2, var, resample=True)

    if ignore_nan:
        b = np.isnan(dI)
        wdiff, dI = wdiff[~b], dI[~b]
        b = np.isnan(I)
        w, I = w[~b], I[~b]
        
    if var in ['transmittance', 'transmittance_noslit']:
        norm = 1 - np.trapz(I, w)
    else:
        norm = np.trapz(I, w)

    return np.abs(np.trapz(dI, wdiff)) / norm

#    b1 = (I_avg == 0)
#    b2 = (dI == 0)

#    with catch_warnings():
#        filterwarnings('ignore', 'invalid value encountered in true_divide')
#        res = dI/I_avg

#    res[b1*b2] = 0             # if both I_avg and dI = 0, solve nan as 0

#    return np.sqrt((res**2).sum())/len(res)
#    return res.mean()


def _get_defaults(s1, s2, var, wunit='default', Iunit='default', medium='default',
                  assert_same_wavelength=False):
    ''' See get_distance, get_diff '''

    # Check inputs, get defaults
    # ----
    if Iunit == 'default':
        try:
            Iunit = s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError('Iunit not defined in spectrum. Cant plot')
    # Format units
    if wunit == 'default':
        wunit = s1.get_waveunit()
    wunit = cast_waveunit(wunit)
    if medium == 'default':
        medium = s1.conditions.get('medium', None)

    # Get data
    # ----
    w1, I1 = s1.get(var, wunit=wunit, Iunit=Iunit, medium=medium)
    w2, I2 = s2.get(var, wunit=wunit, Iunit=Iunit, medium=medium)

    if assert_same_wavelength:
        if not array_allclose(w1, w2):
            raise AssertionError(
                'Wavespace are not the same: use Spectrum.resample()')

    return w1, I1, w2, I2


def plot_diff(s1, s2, var=None, wunit='default', Iunit='default', medium='default',
              resample=True, method='diff',  show_points=False,
              label1=None, label2=None, figsize=None, title=None, nfig=None,
              normalize=False, verbose=True):
    ''' Plot two spectra, and the difference between them

    If waveranges dont match, ``s2`` is interpolated over ``s1``. 


    Parameters    
    ----------

    s1, s2: Spectrum objects

    var: str, or None
        spectral quantity to plot (ex: ``'abscoeff'``). If None, 
        plot the first one in the Spectrum from ``'radiance'``, 
        ``'radiance_noslit'``, ``'transmittance'``, etc.

    wunit: ``'default'``, ``'nm'``, ``'cm-1'``
        if ``'default'``, use first spectrum wunit

    Iunit: str
        if ``'default'``, use first spectrum unit

    medium: ``'air'``, ``'vacuum'``, ``'default'``
        if ``'default'``, use first spectrum propagating medium

    method: ``'distance'``, ``'diff'``, ``'ratio'``
        If ``'diff'``, plot difference at same wavespace position. 
        If ``'distance'``, plot Euclidian distance (note that units are meaningless then)
        If ``'ratio'``, plot ratio of two spectra
        Default ``'diff'``.

        .. warning::
            with ``'distance'``, calculation scales as ~N^2 with N the number
            of points in a spectrum (against ~N with ``'diff'``). This can quickly 
            override all memory.

    normalize: bool
        Normalize the spectra to be ploted 

    Other Parameters
    ----------------

    show_points: boolean
        if True, make all points appear with 'o'

    label1, label2: str
        curve names

    figsize
        figure size

    nfig: int, str
        figure number of name

    title: str
        title

    verbose: boolean
        if True, plot stuff such as rescale ratio in normalize mode. Default ``True``


    Examples
    --------

    ::

        Punit = 'mW/cm2/sr'
        fig, axes = plot_diff(s10, s50, figsize=(18,6),
              label1='brd 10 cm-1, P={0:.2f} {1}'.format(s10.get_power(unit=Punit),Punit),
              label2='brd 50 cm-1, P={0:.2f} {1}'.format(s50.get_power(unit=Punit),Punit)
              )



    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.get_residual`, 
    :meth:`~radis.spectrum.spectrum.compare_with` 

    '''

    # Get defaults
    # ---
    if var is None:    # if nothing is defined, try these first:
        params = s1.get_vars()
        if 'radiance' in params:
            var = 'radiance'
        elif 'radiance_noslit' in params:
            var = 'radiance_noslit'
        elif 'transmittance' in params:
            var = 'transmittance'
        elif 'transmittance_noslit' in params:
            var = 'transmittance_noslit'
        else:
            # or plot the first variable we find
            var = list(params)[0]
            if var.replace('_noslit', '') in params:
                var = var.replace('_noslit', '')
    if Iunit == 'default':
        try:
            Iunit = s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError('Iunit not defined in spectrum. Cant plot')
    if wunit == 'default':
        wunit = s1.get_waveunit()
    if medium == 'default':
        medium = s1.conditions.get('medium', None)

    # Get data
    # ----
    if normalize:
        w1, I1 = s1.get(var, wunit, Iunit, medium)
        w2, I2 = s2.get(var, wunit, Iunit, medium)
        if method == 'distance':
            wdiff, Idiff = curve_distance(w1, I1/np.max(I1), w2, I2/np.max(I2), discard_out_of_bounds=True)
        elif method == 'diff':
            wdiff, Idiff = curve_substract(w1, I1/np.max(I1), w2, I2/np.max(I2))
        elif method == 'ratio':
            wdiff, Idiff = curve_divide(w1, I1/np.max(I1), w2, I2/np.max(I2))
        else:
            raise ValueError('Unknown comparison method: {0}'.format(method))
    else:
        if method == 'distance':
            wdiff, Idiff = get_distance(
                s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
        elif method == 'diff':
            wdiff, Idiff = get_diff(
                s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
        elif method == 'ratio':
            wdiff, Idiff = get_ratio(
                s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
        else:
            raise ValueError('Unknown comparison method: {0}'.format(method))

    # Plot
    # ----

    # Format units
    wunit = cast_waveunit(wunit)
    if wunit == 'cm-1':
        xlabel = 'Wavenumber (cm-1)'
    elif wunit == 'nm':
        xlabel = 'Wavelength (nm)'

    # Init figure
    set_style('origin')
    fig = plt.figure(num=nfig, figsize=figsize)
    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
    ax0 = plt.subplot(gs[0])
    ax1 = plt.subplot(gs[1])
    ax1.get_shared_x_axes().join(ax0, ax1)

    # Plotting style
    if show_points:
        style = '-o'
    else:
        style = '-'

    # Get labels and names
    if label1 is None:
        label1 = s1.get_name()
    if label2 is None:
        label2 = s2.get_name()
    # Plot compared spectra
    if normalize:
        w1, I1 = s1.get(var, wunit, Iunit, medium)
        w2, I2 = s2.get(var, wunit, Iunit, medium)
        if verbose:
            print(('Rescale factor: '+str(np.max(I1)/np.max(I2))))
        ax0.plot(w1, I1/np.max(I1), ls=style, color='k', lw=3, label=label1)
        ax0.plot(w2, I2/np.max(I2), ls=style, color='r', lw=1, label=label2)
    else:
        ax0.plot(*s1.get(var, wunit, Iunit, medium),
                 ls=style, color='k', lw=3, label=label1)
        ax0.plot(*s2.get(var, wunit, Iunit, medium),
                 ls=style, color='r', lw=1, label=label2)

    Iunit = make_up(Iunit)  # cosmetic changes

    ax0.tick_params(labelbottom='off')
    if label1 is not None or label2 is not None:
        ax0.legend(loc='best')

    # Start to 0
    if var in ['radiance_noslit', 'radiance', 'abscoeff', 'absorbance']:
        ax0.set_ylim(bottom=0)

    # plot difference (sorted)
    b = np.argsort(wdiff)
    ax1.plot(wdiff[b], Idiff[b], style, color='k', lw=1)

    if method == 'diff':
        fig.text(0.09, 0.38, 'diff')
    elif method == 'distance':
        fig.text(0.09, 0.38, 'distance')
    elif method == 'ratio':
        fig.text(0.09, 0.38, 'ratio')

    # Write labels
    ax1.set_xlabel(make_up(xlabel))
    if normalize:
        fig.text(0.02, 0.5, 'Arb. Units',
                 va='center', rotation='vertical')
    else:
        fig.text(0.02, 0.5, ('{0} ({1})'.format(make_up(var), Iunit)),
                 va='center', rotation='vertical')

    # Set limits
    if method == 'diff':
        # symmetrize error scale:
        ymin, ymax = ax1.get_ylim()
        ymax = max(abs(ymin), abs(ymax))
        ax1.set_ylim(-ymax, ymax)
    elif method == 'distance':
        ax1.set_ylim(bottom=0)
    elif method == 'ratio':
        # auto-zoom on min, max, but discard first decile (case of spikes / divergences)
        Idiff_s = np.sort(Idiff)
        ax1.set_ylim(bottom=0, top=Idiff_s[len(
            Idiff_s)//10]+Idiff_s[-len(Idiff_s)//10])

    if title:
        fig.suptitle(title)

    # Fix format
    fix_style('origin', ax=ax0)
    fix_style('origin', ax=ax1)
    plt.tight_layout()
    if title:
        plt.subplots_adjust(left=0.15, top=0.92)
    else:
        plt.subplots_adjust(left=0.15)

    # Add cursors
    fig.cursors = MultiCursor(fig.canvas, (ax0, ax1),
                              color='r', lw=1, alpha=0.2, horizOn=False,
                              vertOn=True)

    plt.show()

    return fig, [ax0, ax1]


''' Return the average distance between two spectra.
    It's important to note that if averageDistance(s1, s2)==0 then s1 = s2
    

   .. math::

      \\sqrt{\\sum {(u_i-v_i)^2 / V[x_i]}}.
'''


def averageDistance(s1, s2, var='radiance'):
    ''' Return the average distance between two spectra.
    It's important to note for fitting that if averageDistance(s1, s2)==0 
    then s1 = s2

    Parameters    
    ----------

    s1, s2: Spectrum objects
        spectra to be compared

    var: str, optional
        spectral quantity (ex: 'radiance', 'transmittance'...)

    Returns    
    -------

    distance: float
        Average distance as in the following equation:

        .. math::

            dist = \\frac{\\sqrt{\\sum_i {(s1_i-s2_i)^2}}}{N}.

    '''
    distArray = get_diff(s1, s2, var)
    distArray_Y2 = distArray[1]*distArray[1]
    N = np.size(distArray[1])
    distance = np.sqrt(np.sum(distArray_Y2))/N
    return distance


if __name__ == '__main__':

    from radis.test.spectrum.test_compare import _run_testcases
    print(('Test compare.py: ', _run_testcases(verbose=True)))



# -*- coding: utf-8 -*-
""" Tools to test RADIS library

Summary
-------

Tools to test RADIS library

Todo
---- 

add all files in a separetaly downloadble folder. getTestFile should check
the file exist else suggest to download the folder

Examples
--------

Run all tests::
    
    cd radis/test
    pytest
    
Run only "fast" tests (tests that have a "fast" label, and should be 
a few seconds only)::
    
    cd radis/test
    pytest -m fast

-------------------------------------------------------------------------------

"""

from __future__ import print_function, absolute_import, division, unicode_literals

import os
from radis.misc.config import (getDatabankList, getDatabankEntries, addDatabankEntries,
                               diffDatabankEntries)
from radis.misc.utils import FileNotFoundError
from radis.misc.printer import printr
from os.path import join, dirname

TEST_FOLDER_PATH = join(dirname(dirname(__file__)), 'test')
IGNORE_MISSING_DATABASES = False        # TODO: move in ~/.radis
'''bool: how to deal with missing Line databases during tests. If ``True``, 
print a warning. Else, raise an error. See :func:`~radis.test.utils.IgnoreMissingDatabase`
'''


def getTestFile(file):
    ''' Return the full path of a test file. Used by test functions not to
    worry about the project architecture'''

    return join(TEST_FOLDER_PATH, 'files', file)


def getValidationCase(file):
    ''' Return the full path of a validation case file. Used by test functions not to
    worry about the project architecture'''

    return join(TEST_FOLDER_PATH, 'validation', file)


try:  # Python 3.6 only
    getTestFile.__annotations__['file'] = os.listdir(
        join(TEST_FOLDER_PATH, 'files'))
    getValidationCase.__annotations__['file'] = os.listdir(
        join(TEST_FOLDER_PATH, 'validation'))
except:
    pass

# %% Comparison functions


def testEqual(a, b, info=''):
    if a != b:
        print('Mismatch', info, ':', a, '!=', b)
    return a == b

# %% Test Databases


TEST_DATABASES = {
    'HITRAN-CO2-TEST': {
        'info': 'HITRAN 2016 database, CO2, 1 main isotope (CO2-626), bandhead: ' +
        '2380-2398 cm-1 (4165-4200 nm)',
        'path': [getTestFile(r'hitran_co2_626_bandhead_4165_4200nm.par')],
        'format': 'hitran',
        'parfuncfmt': 'hapi',
        'levelsfmt': 'neq',  # TODO: replace with 'radis'
    },
    'HITRAN-CO-TEST': {
        'info': 'HITRAN 2016 database, CO, 3 main isotopes (CO-26, 36, 28), ' +\
        '2000-2300 cm-1',
        'path': [getTestFile(r'hitran_co_3iso_2000_2300cm.par')],
        'format': 'hitran',
        'parfuncfmt': 'hapi',
        'levelsfmt': 'neq',  # TODO: replace with 'radis'
    },
}

# %% Utils to test spec module


def setup_test_line_databases(verbose=True):
    ''' Build test databases and add them in ~/.radis. Generate the file if it 
    doesnt exist

    In particular:

    - HITRAN-CO2-TEST: CO2, HITRAN 2016, 4165-4200 nm 
    - HITRAN-CO-TEST: CO, HITRAN 2016, 2000-2300 cm-1

    These test databases are used to run the different test routines. They can
    obviously be used by Users to run simulations, but we suggest Users to download
    their own line databases files and add them to ~/.radis so they have more control
    on it

    '''

    # Get list of databases
    try:
        dbnames = getDatabankList()
    except FileNotFoundError:
        dbnames = []

    # %% Add test databases

    def add_to_parser(config, name, dic):
        for k, v in dic.items():
            config[name][k] = v
        if verbose:
            print("Adding '{0}' database in ~/.radis".format(name))

    for dbname, dbentries in TEST_DATABASES.items():

        if dbname in dbnames:  # Check entries are correct
            #            for k
            diff = diffDatabankEntries(getDatabankEntries(dbname), dbentries,
                                       verbose=False)
            if diff is not None:
                raise ValueError('{0}'.format(diff) +
                                 '\nIn ~/.radis\n----------\n{0}'.format(getDatabankEntries(dbname)) +
                                 '\n\nExpected\n---------\n{0}\n\n'.format(dbentries) +
                                 'Test Database {0} doesnt match expected '.format(dbname) +
                                 'entries for key `{0}`. See comparison above. '.format(diff) +
                                 'To regenerate test databases just delete the {0} '.format(dbname) +
                                 'entry in your ~/.radis')

        else:  # add them (create ~/.radis file if doesnt exist yet)
            addDatabankEntries(dbname, dbentries)

    return


# %% Deal with missing databases
def _failsafe_if_no_db(testcase, *args, **kwargs):
    '''finally not implemented?'''
    from radis.misc.utils import DatabankNotFound
    try:
        testcase(*args, **kwargs)
    except DatabankNotFound:
        import sys
        print((sys.exc_info()))
        print(('Testing {0}: Database not defined. \n'.format(testcase.__name__) +
               'Ignoring the test'))
        return True


def IgnoreMissingDatabase(err, file='', warnings=True):
    ''' A function to deal with MissingDatabases errors. If :data:`~radis.test.utils.IGNORE_MISSING_DATABASES`
    is ``True``, just print a warning. Else, raise the error
    
    Parameters
    ----------
    
    err: an Error
    
    file: str
        where the error occured. Use ``file=__file__`` on function call
    '''
    # TODO: make IGNORE_MISSING_DATABASES a ~/.radis parameter
    if IGNORE_MISSING_DATABASES:
        if warnings:
            import sys
            print(sys.exc_info())
            printr('In {0}: Database not defined: {1}'.format(file, err.filename) +
                  '\n Ignoring the test')
        return True
    else:
        raise err


if __name__ == '__main__':
    #    run_tests()
    setup_test_line_databases()



# -*- coding: utf-8 -*-
'''
Summary
-------

HITRAN database parser 


Routine Listing
---------------

- :func:`~radis.io.hitran.hit2df`
- :func:`~radis.io.hitran.get_molecule`
- :func:`~radis.io.hitran.get_molecule_identifier`
- :func:`~radis.io.hitran.parse_local_quanta`
- :func:`~radis.io.hitran.parse_global_quanta`


-------------------------------------------------------------------------------


'''

from __future__ import print_function, absolute_import, division, unicode_literals

import sys
import pandas as pd
from collections import OrderedDict
from os.path import exists, splitext
import radis
from radis.io.tools import parse_binary_file, drop_object_format_columns, replace_PQR_with_m101
from radis.misc.cache_files import save_to_hdf, check_cache_file, get_cache_file


# %% Hitran groups and classes
# As defined in Rothman et al, "The HITRAN 2004 molecular spectroscopic database"
# Tables 3 and 4

# Groups (define local quanta)

HITRAN_GROUP1 = ['H2O', 'O3', 'SO2', 'NO2', 'HNO3', 'H2CO', 'HOCl', 'H2O2', 'COF2',
                 'H2S', 'HO2', 'HCOOH', 'ClONO2', 'HOBr', 'C2H4']
'''str: asymmetric rotors'''

HITRAN_GROUP2 = ['CO2', 'N2O', 'CO', 'HF', 'HCl', 'HBr', 'HI', 'OCS', 'N2', 'HCN',
                 'C2H2', 'NO+']
'''str: diatomic and linear molecules'''

HITRAN_GROUP3 = ['SF6', 'CH4']
'''str: Spherical rotors'''

HITRAN_GROUP4 = ['CH3D', 'CH3Cl', 'C2H6', 'NH3', 'PH3', 'CH3OH']
'''str: symmetric rotors'''

HITRAN_GROUP5 = ['O2']
'''str: Triplet-Sigma ground electronic states'''

HITRAN_GROUP6 = ['NO', 'OH', 'ClO']
'''str: Doublet-Pi ground electronic states'''

# Classes (define global quanta)

HITRAN_CLASS1 = ['CO', 'HF', 'HCl', 'HBr', 'HI', 'N2', 'NO+']
'''str: Diatomic molecules with ? '''

HITRAN_CLASS2 = ['O2']
'''str: Diatomic molecules with different electronic levels'''

HITRAN_CLASS3 = ['NO', 'OH', 'ClO']
'''str: Diatomic molecules with doublet-Pi electronic state'''

HITRAN_CLASS4 = ['N2O', 'OCS', 'HCN']
'''str: Linear triatomic'''

HITRAN_CLASS5 = ['CO2']
'''str: Linear triatomic with large Fermi resonance'''

HITRAN_CLASS6 = ['H2O', 'O3', 'SO2', 'NO2', 'HOCl', 'H2S', 'HO2', 'HOBr']
'''str: Non-linear triatomic'''

HITRAN_CLASS7 = ['C2H2']
'''str: Linear tetratomic'''

HITRAN_CLASS8 = ['NH3', 'PH3']
'''str: Pyramidal tetratomic'''

HITRAN_CLASS9 = ['H2CO', 'H2O2', 'COF2']
'''str: Non-linear tetratomic'''

HITRAN_CLASS10 = ['CH4', 'CH3D', 'CH3Cl', 'C2H6', 'HNO3', 'SF6', 'HCOOH',
                  'ClONO2', 'C2H4', 'CH3OH']
'''str: Pentatomic or greater polyatomic'''

# %% Parsing functions

# General case
columns_2004 = OrderedDict([(
    # name    # format # type  # description                                 # unit
    'id',     ('a2',   int,   'Molecular number', '')), (
    'iso',    ('a1',   int,   'isotope number', '')), (
    'wav',    ('a12',  float, 'vacuum wavenumber', 'cm-1')), (
    'int',    ('a10',  float, 'intensity at 296K', 'cm-1/(molecule/cm-2)', )), (
    'A',      ('a10',  float, 'Einstein A coefficient', 's-1')), (
    'airbrd', ('a5',   float, 'air-broadened half-width at 296K', 'cm-1.atm-1')), (
    'selbrd', ('a5',   float, 'self-broadened half-width at 296K', 'cm-1.atm-1')), (
    'El',     ('a10',  float, 'lower-state energy', 'cm-1')), (
    'Tdpair', ('a4',   float, 'temperature-dependance exponent for Gamma air', '')), (
    'Pshft',  ('a8',   float, 'air pressure-induced line shift at 296K', 'cm-1.atm-1')), (
    'globu',  ('a15',  str,   'electronic and vibrational global upper quanta', '')), (
    'globl',  ('a15',  str,   'electronic and vibrational global lower quanta', '')), (
    'locu',   ('a15',  str,   'electronic and vibrational local upper quanta', '')), (
    'locl',   ('a15',  str,   'electronic and vibrational local lower quanta', '')), (
    'ierr',   ('a6',   str,   'ordered list of indices corresponding to uncertainty estimates of transition parameters', '')), (
    'iref',   ('a12',  str,   'ordered list of reference identifiers for transition parameters', '')), (
    'lmix',   ('a1',   str,   'flag indicating the presence of additional data and code relating to line-mixing', '')), (
    'gp',     ('a7',   float, 'upper state degeneracy', '')), (
    'gpp',    ('a7',   float, 'lower state degeneracy', ''))
])


def hit2df(fname, count=-1, cache=False, verbose=True, drop_non_numeric=True):
    ''' Convert a HITRAN/HITEMP [1]_ file to a Pandas dataframe 

    Parameters    
    ----------

    fname: str
        HITRAN-HITEMP file name 

    count: int
        number of items to read (-1 means all file)

    cache: boolean, or ``'regen'`` or ``'force'``
        if ``True``, a pandas-readable HDF5 file is generated on first access, 
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not 
        taken into account). If False, no database is used. If ``'regen'``, temp
        file are reconstructed. Default ``False``. 

    Other Parameters
    ----------------
    
    drop_non_numeric: boolean
        if ``True``, non numeric columns are dropped. This improves performances, 
        but make sure all the columns you need are converted to numeric formats 
        before hand. Default ``True``. Note that if a cache file is loaded it 
        will be left untouched.

    Returns
    -------

    df: pandas Dataframe
        dataframe containing all lines and parameters



    References
    ----------


    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__



    Notes
    -----

    Performances: see CDSD-HITEMP parser


    See Also
    --------
    
    :func:`~radis.io.cdsd.cdsd2df`

    '''

    columns = columns_2004

    # Use cache file if possible
    fcache = splitext(fname)[0]+'.h5'
    check_cache_file(cache, fcache=fcache, verbose=verbose)
    if cache and exists(fcache):
        return get_cache_file(fcache)

    # Detect the molecule by reading the start of the file
    with open(fname) as f:
        mol = get_molecule(int(f.read(2)))

    # %% Start reading the full file

    df = parse_binary_file(fname, columns, count)

    # %% Post processing

    # assert one molecule per database only. Else the groupbase data reading
    # above doesnt make sense
    nmol = len(set(df['id']))
    if nmol == 0:
        raise ValueError('Databank looks empty')
    elif nmol != 1:
        # Crash, give explicity error messages
        try:
            secondline = df.iloc[1]
        except IndexError:
            secondline = ''
        raise ValueError('Multiple molecules in database ({0}). Current '.format(nmol) +
                         'spectral code only computes 1 species at the time. Use MergeSlabs. ' +
                         'Verify the parsing was correct by looking at the first row below: ' +
                         '\n{0}'.format(df.iloc[0])+'\n----------------\nand the second row ' +
                         'below: \n{0}'.format(secondline))

    # dd local quanta attributes, based on the HITRAN group
    df = parse_local_quanta(df, mol)

    # Add global quanta attributes, based on the HITRAN class
    df = parse_global_quanta(df, mol)
    
    # Remove non numerical attributes
    if drop_non_numeric:
        if 'branch' in df:
            replace_PQR_with_m101(df)
        df = drop_object_format_columns(df, verbose=verbose)

    # cached file mode but cached file doesn't exist yet (else we had returned)
    if cache:
        if verbose:
            print('Generating cached file: {0}'.format(fcache))
        try:
            save_to_hdf(df, fcache, metadata={}, version=radis.__version__,
                        key='df', overwrite=True)
        except:
            if verbose:
                print(sys.exc_info())
                print('An error occured in cache file generation. Lookup access rights')
            pass

    return df


# %% Hitran global quanta classes

def _parse_HITRAN_class1(df):
    ''' Diatomic molecules: CO, HF, HCl, HBr, HI, N2, NO+


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_ :

    >>>       v
    >>>  13x I2
    
    References
    ----------
    
    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    '''
    
    # 1. Parse
    dgu = df['globu'].str.extract(
        '[ ]{13}(?P<vu>[\d ]{2})',
        expand=True)
    dgl = df['globl'].str.extract(
        '[ ]{13}(?P<vl>[\d ]{2})',
        expand=True)
    
    # 2. Convert to numeric
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    
    # 3. Clean
    del df['globu']
    del df['globl']
    
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class2(df):
    ''' Diatomic molecules with different electronic levels: O2


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_ :

    >>>      X   v
    >>>  12x A1 I2

    References
    ----------
    
    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    print('parse_global_quanta not implemented for molecules of HITRAN class 2')
    return df


def _parse_HITRAN_class3(df):
    ''' Diatomic molecules with doublet-Pi electronic state: NO, OH, ClO


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -------

    HITRAN syntax [1]_:

    >>>      X i     v1
    >>>  7x A1 A3 2x I2
    
    References
    ----------
    
    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    print('parse_global_quanta not implemented for molecules of HITRAN class 3')
    return df


def _parse_HITRAN_class4(df):
    ''' Parse linear triatomic class in HITRAN [1]_: N2O, OCS, HCN

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 l2 v3
    >>>  7x I2 I2 I2 I2

    Note: I2 in regexp: [\d ]{2} 

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    
    # 1. Parse
    dgu = df['globu'].str.extract(
        '[ ]{7}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})',
        expand=True)
    dgl = df['globl'].str.extract(
        '[ ]{7}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})',
        expand=True)
    
    # 2. Convert to numeric
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    
    # 3. Clean
    del df['globu']
    del df['globl']
    
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class5(df):
    ''' Parse linear triatomic with large Fermi resonance in HITRAN [1]_: CO2

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 l2 v3 r
    >>>  6x I2 I2 I2 I2 I1

    Note: I2 in regexp: [\d ]{2} 

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    
    # 1. Parse
    dgu = df['globu'].str.extract(
        '[ ]{6}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})(?P<ru>\d)',
        expand=True)
    dgl = df['globl'].str.extract(
        '[ ]{6}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})(?P<rl>\d)',
        expand=True)
    
    # 2. Convert to numeric
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    
    # 3. Clean
    del df['globu']
    del df['globl']
    
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class6(df):
    ''' Parse non-linear triatomic in HITRAN [1]_: H2O, O3, SO2, NO2, HOCl, H2S, HO2, HOBr

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 v3
    >>>  9x I2 I2 I2

    Note: I2 in regexp: [\d ]{2} 

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    
    # 1. Parse
    dgu = df['globu'].str.extract(
#        '[ ]{9}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<v3u>[\d ]{2})',
        '[ ]{9}(?P<v1u>[\-\d ]{2})(?P<v2u>[\-\d ]{2})(?P<v3u>[\-\d ]{2})',
        expand=True)
    dgl = df['globl'].str.extract(
#        '[ ]{9}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<v3l>[\d ]{2})',
        '[ ]{9}(?P<v1l>[\-\d ]{2})(?P<v2l>[\-\d ]{2})(?P<v3l>[\-\d ]{2})',
        expand=True)
    # ... note @EP: in HITRAN H2O files, for iso=2, vibrational levels are 
    # ... somehow negative. The regex above is adapted to catch negation signs with \-
    
    # 2. Convert to numeric
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    
    # 3. Clean
    del df['globu']
    del df['globl']
    
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class7(df):
    ''' Parse linear tetratomic in HITRAN [1]_: C2H2

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    print('parse_global_quanta not implemented for molecules of HITRAN class 7')
    return df


def _parse_HITRAN_class8(df):
    ''' Pyramidal tetratomic in HITRAN [1]_: NH3, PH3


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    print('parse_global_quanta not implemented for molecules of HITRAN class 8')
    return df


def _parse_HITRAN_class9(df):
    ''' Non-linear tetratomic in HITRAN [1]_: H2CO, H2O2, COF2


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    print('parse_global_quanta not implemented for molecules of HITRAN class 9')
    return df


def _parse_HITRAN_class10(df):
    ''' Pentatomic or greater polyatomic in HITRAN [1]_


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    '''
    print('parse_global_quanta not implemented for molecules of HITRAN class 10')
    return df

# %% HITRAN Local quanta


def _parse_HITRAN_group1(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax: [1]_


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    '''
    
    # 1. Parse
    
    # Ref [1] : locu
    # --------------
    # J'  | Ka' | Kc' | F'  | Sym'
    # I3  | I3  | I3  | A5  | A1
    dgu = df['locu'].str.extract(
        '(?P<ju>[\d ]{3})(?P<Kau>[\-\d ]{3})(?P<Kcu>[\-\d ]{3})(?P<Fu>.{5})(?P<symu>.)',
        expand=True)
    # Ref [1] : locl
    # --------------
    # J'' | Ka''| Kc''| F'' | Sym''
    # I3  | I3  | I3  | A5  | A1
    dgl = df['locl'].str.extract(
        '(?P<jl>[\d ]{3})(?P<Kal>[\-\d ]{3})(?P<Kcl>[\-\d ]{3})(?P<Fl>.{5})(?P<syml>.)',
        expand=True)
    # ... note @EP: in HITRAN H2O files, for iso=2, the Kau, Kcu can somehow
    # ... be negative. The regex above is adapted to catch negation signs with \-

    # 2. Convert to numeric
    for k in ['ju', 'Kau', 'Kcu']:
        dgu[k] = pd.to_numeric(dgu[k])
    for k in ['jl', 'Kal', 'Kcl']:
        dgl[k] = pd.to_numeric(dgl[k])
    
    # 3. Clean
    del df['locu']
    del df['locl']
    
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_group2(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax: [1]


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    '''

    # 1. Parse
    
    # Ref [1] : locu
    # --------------
    #     | F'  |
    # 10X | A5  |
    dgu = df['locu'].str.extract(
        '[ ]{10}(?P<Fu>.{5})',
        expand=True)
    # Ref [1] : locl
    # --------------
    #     | Br  | J'' | Sym''| F'' |
    # 5X  | A1  | I3  | A1   | A5  |
    dgl = df['locl'].str.extract(
        '[ ]{5}(?P<branch>[\S]{1})(?P<jl>[\d ]{3})(?P<syml>.)(?P<Fl>.{5})',
        expand=True)

    # 2. Convert to numeric
    
    #dgl['jl'] = dgl.jl.apply(pd.to_numeric)
    dgl['jl'] = pd.to_numeric(dgl.jl)

    # 3. Clean
    del df['locu']
    del df['locl']

    return pd.concat([df, dgu, dgl], axis=1)


#                # 10X included in Fu
#               'Fu',     ('a15',  str,  'upper state total angular momentum including nuclear spin'  ,''         )),(
#               #'locl',   ('a15',  str,   'electronic and vibrational local lower quanta'  ,''                      )),(
#               'branch', ('a6',   str,     'O, P, Q, R, S branch symbol'                  ,''                      )),(
#               'jl',     ('a3',   int,    'lower state rotational quantum number'         ,''                      )),(
#               'sym',    ('a1',   str,     'symmetry'                                     ,''                      )),(
#               'Fl',     ('a5',   str,   'lower state total angular momentum including nuclear spin', ''         )),(


def _parse_HITRAN_group3(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    '''
    print('parse_local_quanta not implemented for molecules of HITRAN group 3')
    return df


def _parse_HITRAN_group4(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    '''
    print('parse_local_quanta not implemented for molecules of HITRAN group 4')
    return df


def _parse_HITRAN_group5(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    '''
    print('parse_local_quanta not implemented for molecules of HITRAN group 5')
    return df


def _parse_HITRAN_group6(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    '''
    print('parse_local_quanta not implemented for molecules of HITRAN group 6')
    return df


# %% Reading function


def parse_local_quanta(df, mol):
    '''
    Parameters
    ----------

    df: pandas Dataframe

    mol: str
        molecule name
    '''

    # had some problems with bytes types
    df['locu'] = df.locu.astype(str)
    df['locl'] = df.locl.astype(str)

    if mol in HITRAN_GROUP1:
        df = _parse_HITRAN_group1(df)
    elif mol in HITRAN_GROUP2:
        df = _parse_HITRAN_group2(df)
    elif mol in HITRAN_GROUP3:
        df = _parse_HITRAN_group3(df)
    elif mol in HITRAN_GROUP4:
        df = _parse_HITRAN_group4(df)
    elif mol in HITRAN_GROUP5:
        df = _parse_HITRAN_group5(df)
    elif mol in HITRAN_GROUP6:
        df = _parse_HITRAN_group6(df)
    else:
        raise ValueError('Unknown group for molecule {0}. Cant parse local quanta'.format(
            mol))

    return df


def parse_global_quanta(df, mol):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe

    mol: str
        molecule name
    '''

    # had some problems with bytes types
    df['globu'] = df.globu.astype(str)
    df['globl'] = df.globl.astype(str)

    if mol in HITRAN_CLASS1:
        df = _parse_HITRAN_class1(df)
    elif mol in HITRAN_CLASS2:
        df = _parse_HITRAN_class2(df)
    elif mol in HITRAN_CLASS3:
        df = _parse_HITRAN_class3(df)
    elif mol in HITRAN_CLASS4:
        df = _parse_HITRAN_class4(df)
    elif mol in HITRAN_CLASS5:
        df = _parse_HITRAN_class5(df)
    elif mol in HITRAN_CLASS6:
        df = _parse_HITRAN_class6(df)
    elif mol in HITRAN_CLASS7:
        df = _parse_HITRAN_class7(df)
    elif mol in HITRAN_CLASS8:
        df = _parse_HITRAN_class8(df)
    elif mol in HITRAN_CLASS9:
        df = _parse_HITRAN_class9(df)
    elif mol in HITRAN_CLASS10:
        df = _parse_HITRAN_class10(df)
    else:
        raise ValueError('Unknown class for molecule {0}. Cant parse global quanta'.format(
            mol))

    return df

def get_molecule_identifier(molecule_name):
    '''
    For a given input molecular formula, return the corresponding HITRAN molecule 
    identifier number [1]_.


    Parameters
    ----------
    molecular_formula : str
        The string describing the molecule.


    Returns
    -------
    M : int
        The HITRAN molecular identified number.


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    Function is from from https://github.com/nzhagen/hitran/blob/master/hitran.py

    '''

    trans = {'1': 'H2O',    '2': 'CO2',   '3': 'O3',      '4': 'N2O',
             '5': 'CO',    '6': 'CH4',   '7': 'O2',     '8': 'NO',
             '9': 'SO2',   '10': 'NO2',  '11': 'NH3',    '12': 'HNO3',
             '13': 'OH',   '14': 'HF',   '15': 'HCl',   '16': 'HBr',
             '17': 'HI',    '18': 'ClO',  '19': 'OCS',    '20': 'H2CO',
             '21': 'HOCl', '22': 'N2',   '23': 'HCN',   '24': 'CH3Cl',
             '25': 'H2O2',  '26': 'C2H2', '27': 'C2H6',   '28': 'PH3',
             '29': 'COF2', '30': 'SF6',  '31': 'H2S',   '32': 'HCOOH',
             '33': 'HO2',   '34': 'O',    '35': 'ClONO2', '36': 'NO+',
             '37': 'HOBr', '38': 'C2H4', '39': 'CH3OH', '40': 'CH3Br',
             '41': 'CH3CN', '42': 'CF4',  '43': 'C4H2',   '44': 'HC3N',
             '45': 'H2',   '46': 'CS',   '47': 'SO3'}
    # Invert the dictionary.
    trans = {v: k for k, v in trans.items()}

    return int(trans[molecule_name])


def get_molecule(molecule_id):
    '''
    For a given input molecular identifier, return the corresponding HITRAN 
    molecule name [1]_.


    Parameters    
    ----------

    molecular_id : str
        Hitran identifier of the molecule.


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''

    # assert str
    id = '{0}'.format(molecule_id)

    trans = {'1': 'H2O',    '2': 'CO2',   '3': 'O3',      '4': 'N2O',
             '5': 'CO',    '6': 'CH4',   '7': 'O2',     '8': 'NO',
             '9': 'SO2',   '10': 'NO2',  '11': 'NH3',    '12': 'HNO3',
             '13': 'OH',   '14': 'HF',   '15': 'HCl',   '16': 'HBr',
             '17': 'HI',    '18': 'ClO',  '19': 'OCS',    '20': 'H2CO',
             '21': 'HOCl', '22': 'N2',   '23': 'HCN',   '24': 'CH3Cl',
             '25': 'H2O2',  '26': 'C2H2', '27': 'C2H6',   '28': 'PH3',
             '29': 'COF2', '30': 'SF6',  '31': 'H2S',   '32': 'HCOOH',
             '33': 'HO2',   '34': 'O',    '35': 'ClONO2', '36': 'NO+',
             '37': 'HOBr', '38': 'C2H4', '39': 'CH3OH', '40': 'CH3Br',
             '41': 'CH3CN', '42': 'CF4',  '43': 'C4H2',   '44': 'HC3N',
             '45': 'H2',   '46': 'CS',   '47': 'SO3'}

    return trans[id]

# ======================================================
# %% Test


if __name__ == '__main__':
    from radis.test.test_io import _run_testcases
    print('Testing HITRAN parsing: ', _run_testcases())



# -*- coding: utf-8 -*-
"""
Summary
-------

Runs tests for neq/phys so that they can be accessed by pytest (and hopefully 
the CI test suite)

Examples
--------

Run all tests::

    pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)::

    pytest -m fast

-------------------------------------------------------------------------------

"""


from __future__ import absolute_import, unicode_literals, print_function
import numpy as np
from numpy import isclose
from radis.phys.convert import (J2eV, J2cm, cm2J, eV2cm, eV2K, eV2nm, nm2eV,
                                dnm2dcm, nm2cm, dnm2dhz, dhz2dnm, cm2eV, eV2J, K2J,
                                J2K, dcm2dnm, K2eV, hz2nm, nm2hz,
                                torr2bar, torr2atm, bar2torr, bar2atm, atm2torr,
                                atm2bar)
from radis.phys.units import uarray, Q_, conv2
import pytest


@pytest.mark.fast
def test_convert(verbose=True, *args, **kwargs):
    ''' Test conversions  '''

    E = np.linspace(1, 5, 5)  # eV
    assert (J2eV(K2J(J2K(eV2J(E)))) == E).all()

    E = 2150  # cm-1
    assert J2cm(cm2J(E))

    E = 1  # eV
    assert (eV2cm(E) == J2cm(eV2J(1)))
    assert (round(eV2K(E), 0) == 11605)
    assert K2eV(eV2K(E)) == E

    E = 250  # nm
    assert isclose(nm2eV(E), cm2eV(nm2cm(E)))
    assert (eV2nm(nm2eV(E)) == E)
    assert hz2nm(nm2hz(E)) == E

    fwhm = 1.5  # nm
    lbd_0 = 632.8  # nm
    assert (isclose(fwhm, dcm2dnm(dnm2dcm(fwhm, lbd_0), nm2cm(lbd_0))))

    fwhm = 2e-3  # 0.002 nm
    lbd_0 = 632.8
    fwhm_hz = dnm2dhz(fwhm, lbd_0)  # ~ 1.5 GHz
    if verbose:
        print(('{0:.2g} nm broadening at {1} nm = {2:.2g} Ghz'.format(
            fwhm, lbd_0, fwhm_hz*1e-9)))
    assert isclose(fwhm_hz*1e-9, 1.4973307983125002)
    assert isclose(dhz2dnm(fwhm_hz, nm2hz(lbd_0)), fwhm)

    assert atm2bar(1) == 1.01325
    assert isclose(torr2bar(atm2torr(bar2atm(1))), 1)
    assert isclose(torr2atm(bar2torr(atm2bar(1))), 1)

    return True


def test_units(verbose=True, *args, **kwargs):

    # Test unit-ware arrays
    # RADIS pint-aware array
    a = uarray(np.linspace(10, 100, 10), 'Td')
    res = Q_(np.linspace(1e-16, 1e-15, 10), 'V * cm^2')  # pint definition

    assert (np.round(np.array(a.to('V * cm^2')) - np.array(res), 5)
            == np.zeros_like(res)).all()
#    b = (print(a.to('V * cm^2'))==print(res))

    # Test conversion
    convtable = [
        (500, 'nm', 0.5, 'µm'),
        (1, 'erg/s', 1e-7, 'W'),
        (1, 'm2', 10000, 'cm2')
    ]
    for a, f, r, t in convtable:
        cr = conv2(a, f, t)
        if verbose:
            print(('{0} {1} = {2} {3}'.format(a, f, cr, t)))
        assert isclose(cr, r)

    return True


def _run_testcases(*args, **kwargs):

    assert test_convert(*args, **kwargs)
    assert test_units(*args, **kwargs)

    return True


if __name__ == '__main__':
    print('testing phys.py:', _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
""" Test parsers

Notes
-----

Runs tests for radis/io so that they can be accessed by pytest (and hopefully 
the CI test suite)

Examples
--------

Run all tests::

	pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)::

	pytest -m fast

-------------------------------------------------------------------------------

"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.io.hitran import hit2df
from radis.io.cdsd import cdsd2df
from radis.test.utils import getTestFile
import pytest
import numpy as np


@pytest.mark.fast
def test_hitran_co(verbose=True, warnings=True, **kwargs):
    ''' Analyse some default files to make sure everything still works'''

    # 1. Load
    df = hit2df(getTestFile('hitran_CO_fragment.par'), cache='regen')
    if verbose:
        print('Read hitran_CO_fragment.par')
        print('---------------------------')
        print(df.head())
        
    # 2. Test
    assert (list(df.loc[0, ['vu', 'vl']]) == [4, 4])
    assert df.dtypes['vu'] == np.int64
    assert df.dtypes['vl'] == np.int64
    
    return True

def test_hitran_co2(verbose=True, warnings=True, **kwargs):
    
    # 1. Load
    df = hit2df(getTestFile('hitran_CO2_fragment.par'), cache='regen')
    if verbose:
        print('Read hitran_CO2_fragment.par')
        print('----------------------------')
        print(df.head())
        
    # 2. Test
    assert (list(df.loc[0, ['v1u', 'v2u', 'l2u', 'v3u', 'v1l', 'v2l', 'l2l', 'v3l']]) ==
            [4, 0, 0, 0, 0, 0, 0, 1])
    assert df.dtypes['v1l'] == np.int64
    assert df.dtypes['v3u'] == np.int64

    return True

def test_hitran_h2o(verbose=True, warnings=True, **kwargs):
    
    # 1. Load
    df = hit2df(getTestFile('hitran_2016_H2O_2iso_2000_2100cm.par'), cache='regen')
    if verbose:
        print('Read hitran_2016_H2O_2iso_2000_2100cm.par')
        print('-----------------------------------------')
        print(df.head())
        
    # 2. Test
    assert (list(df.loc[0, ['v1u', 'v2u', 'v3u', 'v1l', 'v2l', 'v3l']]) ==
            [0, 2, 0, 0, 1, 0])
    
    assert df.loc[26, 'ju'] == 5    # in .par : line 27, column 99-100
    assert df.loc[27, 'ju'] == 18   # in .par : line 28, column 99-100
    
    
    assert df.dtypes['v1l'] == np.int64
    assert df.dtypes['v3u'] == np.int64
    
    assert df.dtypes['ju'] == np.int64
    assert df.dtypes['Kau'] == np.int64
    assert df.dtypes['Kcu'] == np.int64
    assert df.dtypes['jl'] == np.int64
    assert df.dtypes['Kal'] == np.int64
    assert df.dtypes['Kcl'] == np.int64
    
    return True


def test_hitemp(verbose=True, warnings=True, **kwargs):
    ''' Analyse some default files to make sure everything still works'''

    # 1. Load
    df = cdsd2df(getTestFile('cdsd_hitemp_09.txt'), cache='regen',
                 drop_non_numeric=True)
    if verbose:
        print(df.head())
    
    # 2. Tests
    assert df.wav[3] == 2250.00096
    # make sure P Q R is correctly replaced by drop_non_numeric: 
    assert 'branch' in df    
    assert df['branch'].iloc[0] == 0 # Q
    assert df['branch'].iloc[1] == 1 # R
    
    return True

def _run_testcases(verbose=True, *args, **kwargs):

#    test_hitran_co(verbose=verbose, *args, **kwargs)
#    test_hitran_co2(verbose=verbose, *args, **kwargs)
    test_hitran_h2o(verbose=verbose, *args, **kwargs)
#    test_hitemp(verbose=verbose, *args, **kwargs)

    return True


if __name__ == '__main__':
    print('Testing io.py: ', _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Print functions, with colors
"""

from __future__ import print_function
from termcolor import colored
from six import StringIO
import sys

# %% Colored Print functions


def printm(*args, **kwargs):
    ''' Print in magenta'''

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color='magenta'))


def printr(*args, **kwargs):
    ''' Print in red'''

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color='red'))


def _capture_print(*args, **kwargs):
    ''' Emulate print option but get output in a ```str``` variable instead of stdout '''

    # Change the output to capture the string instead of sending it to the console
    old_stdout = sys.stdout
    sys.stdout = newstdout = StringIO()     # capture all print
    try:
        print(*args, **kwargs)
        string = newstdout.getvalue()              # Get string output
    except:
        raise
    finally:
        sys.stdout = old_stdout

    # discard line return character (will be added again when we print ``string`` )
    return string[:-1]



# -*- coding: utf-8 -*-
"""
Created on Fri Jul  6 13:52:04 2018

@author: erwan
"""

import numpy as np
import pandas as pd
from six.moves import range
from six.moves import zip


def parse_binary_file(fname, columns, count):
    ''' Parse a file under HITRAN ``par`` format. Parsing is done in binary 
    format so it's as fast as possible.

    Parameters
    ----------

    fname: str
        filename

    columns: dict
        list of columns and their format

    count: int
        number of lines to read

    Returns
    -------

    df: pandas DataFrame
        dataframe with lines

    Notes
    -----

    Part common to hit2df and cdsd2df

    '''

    # To be faster, we read file totally in bytes mode with fromfiles. But that
    # requires to properly decode the line return character:

    # problem arise when file was written in an OS and read in another OS (for instance,
    # line return characters are not converted when read from .egg files). Here
    # we read the first line and infer the line return character for it

    # ... Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()]+[('_linereturn', 'a2')]
    # ... _linereturn is to capture the line return symbol. We delete it afterwards
    dt = _format_dtype(dtype)
    data = np.fromfile(fname, dtype=dt, count=1)   # just read the first line

    # get format of line return
    from radis.misc.basics import to_str
    linereturn = to_str(data[0][-1])
    if to_str('\r\n') in linereturn:
        linereturnformat = 'a2'
    elif to_str('\n') in linereturn or to_str('\r') in linereturn:
        linereturnformat = 'a1'
    else:
        raise ValueError(
            'Line return format unknown: {0}. Check your file format.'.format(linereturn))

    # Now re-read with correct line return character

    # ... Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()] + \
        [('_linereturn', linereturnformat)]
    # ... _linereturn is to capture the line return symbol. We delete it afterwards
    dt = _format_dtype(dtype)
    data = np.fromfile(fname, dtype=dt, count=count)

    # ... Cast to new type
    # This requires to recast all the data already read, but is still the fastest
    # method I found to read a file directly (for performance benchmark see
    # CDSD-HITEMP parser)
    newtype = [c[0] if (c[1] == str) else c[1] for c in columns.values()]
    dtype = list(zip(list(columns.keys()), newtype)) + \
        [('_linereturn', linereturnformat)]
    data = _cast_to_dtype(data, dtype)

    # %% Create dataframe
    df = pd.DataFrame(data.tolist(), columns=list(
        columns.keys())+['_linereturn'])

    # Delete dummy column than handled the line return character
    del df['_linereturn']

    # Update format
    for k, c in columns.items():
        if c[1] == str:
            df[k] = df[k].str.decode("utf-8")

    # Strip whitespaces around PQR columns (due to 2 columns jumped)
    if 'branch' in df:  # (only in CDSD)
        df['branch'] = df.branch.str.strip()

    return df


def _format_dtype(dtype):
    ''' Format dtype from specific columns. Crash with hopefully helping error message '''

    try:
        dt = np.dtype([(str(k), c) for k, c in dtype])
        # Note: dtype names cannot be `unicode` in Python2. Hence the str()
    except TypeError:
        # Cant read database. Try to be more explicit for user
        print('Data type')
        print('-'*30)
        for (k, c) in dtype:
            print(str(k), '\t', c)
        print('-'*30)
        raise
    return dt


def _cast_to_dtype(data, dtype):
    ''' Cast array to certain type, crash with hopefull helping error message.
    Return casted data


    Parameters    
    ----------

    data: array to cast

    dtype: (ordered) list of (param, type)

    '''

    dt = _format_dtype(dtype)

    try:
        data = np.array(data, dtype=dt)
    except ValueError:
        try:
            # Cant read database. Try to be more explicit for user
            print('Cant cast data to specific dtype. Trying column by column:')
            print('-'*30)
            for i in range(len(data[0])):
                print(dtype[i], '\t', np.array(data[0][i], dtype=dt[i]))
            print('-'*30)
        except ValueError:
            print('>>> Next param:', dtype[i], '. Value:', data[0][i], '\n')
            raise ValueError(
                'Cant cast data to specific dtype. Tried column by column. See results above')

    return data


def drop_object_format_columns(df, verbose=True):
    ''' Remove 'object' columns in a pandas DataFrame. They are not useful to us at this 
    time, and they slow down all operations (as they are converted to 'object'
    in pandas DataFrame). If you want to keep them, better convert them to 
    some numeric values
    '''
    
    objects = [k for k,v in df.dtypes.items() if v == object]
    for k in objects:
        del df[k]
    if verbose and len(objects)>0:
        print('The following columns had the `object` format and were removed: {0}'.format(
                objects))
    return df
    



def replace_PQR_with_m101(df):
    ''' Return P, Q, R in column ``branch`` with -1, 0, 1 to get a fully numeric 
    database. This improves performances quite a lot, as Pandas doesnt have a 
    fixed-string dtype hence would use the slow ``object`` dtype. 
    
    Parameters
    ----------
    
    df: pandas Dataframe
        ``branch`` must be a column name
    
    Returns
    -------
    
    None:
        ``df`` is is modified in place 
    '''
        
    # Note: somehow pandas updates dtype automatically. We have to check
    
    if df.dtypes['branch'] != np.int64:
        df.loc[df.branch=='P', 'branch'] = -1
    if df.dtypes['branch'] != np.int64:
        df.loc[df.branch=='Q', 'branch'] = 0
    if df.dtypes['branch'] != np.int64:
        df.loc[df.branch=='R', 'branch'] = 1
    
    assert df.dtypes['branch'] == np.int64  



# -*- coding: utf-8 -*-
"""

Notes
-----

Dimensioned calculation is slower:

Performance test with pint:

- 2 times slower than numpy on np.std(np.exp())

Examples
--------

Import neq.phys.units to get access to::

    from neq.phys.units import Q_
    a = Q_(np.array([5,4,2]),'cm')
    a.to_base_units()

-------------------------------------------------------------------------------


"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import numpy as np
from os.path import join, dirname, abspath, exists
from pint import UnitRegistry, DimensionalityError

ureg = UnitRegistry()
_units_file = abspath(join(dirname(__file__), 'units.txt'))
# make sure file exists
if not exists(_units_file):
    raise AssertionError(
        "Couldn't find units file in : {0}".format(_units_file))
ureg.load_definitions(_units_file)
Q_ = ureg.Quantity

# %% Pint aware arrays


class uarray(np.ndarray):
    ''' Unit-aware array based on Pint

    Example
    -------

    >>> a = uarray(np.linspace(10, 100, 10), 'Td')

    '''

    def __new__(cls, input_array, unit=None):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type

        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance

        if unit is not None:
            obj = Q_(obj, unit)

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
#        self.info = getattr(obj, 'info', None)


def conv2(quantity, fromunit, tounit):
    ''' Converts `quantity` from unit `fromunit` to unit `tounit`


    Parameters    
    ----------

    quantity: array
        quantity to convert

    fromunit: str
        input unit

    tounit: str
        output unit


    Note
    ----

    The output is still non dimensional. We don't transform `quantity` 
    into a pint array (or neq.phys.uarray) because this may create a performance
    drop in computationaly-expensive task. Instead, we assume we know for 
    sure the units in which some of our quantities will be created, and just
    want to let the users choose another output unit 

    '''

    try:
        a = Q_(quantity, fromunit)
        a = a.to(tounit)
    except TypeError:
        if 'cm-1' in fromunit or 'cm-1' in tounit:
            #            raise TypeError('Use cm_1 instead of cm-1 else it triggers errors in '+\
            #                            'pint (symbolic unit converter)')
            conv2(quantity, fromunit.replace('cm-1', 'cm_1'),
                  tounit.replace('cm-1', 'cm_1'))

        else:
            raise

    return a.magnitude


def is_homogeneous(unit1, unit2):
    ''' Tells if unit1 and unit2 are homogeneous, using the Pint library


    Parameters    
    ----------

    unit1, unit2: str
        units 

    '''

    try:
        Q_(unit1) + Q_(unit2)
        return True
    except DimensionalityError:
        return False

# %% Emission density conversion functions (with changes from ~1/nm in ~1/cm-1)


def convert_emi2cm(j_nm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral emission density in wavelength base (typically ~mW/cm3/sr/nm) 
    to spectral emission density in wavenumber base (~mW/cm3/sr/cm_1)


    Parameters    
    ----------

    j_nm: array
        spectral emission density in Iunit0 unit (~ per wavelength) )

    wavenum: array (cm-1)
        wavenumber

    Iunit0: str
        unit (~ per wavelength) to convert from 

    Iunit: str
        unit (~ per wavenumber) to convert to


    Notes
    -----

    Implementation:

    We use the variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lν = - Lλ * 1e7 / ν^2


    Example
    -------

    Validation:

        >>> w_nm, j_nm = s.get('emisscoeff', 'nm', 'mW/cm2/sr/nm')
        >>> w_cm, j_cm = s.get('emisscoeff', 'cm', 'mW/cm2/sr/cm_1')
        >>> print(trapz(y_nm, x_nm))
        >>> print(trapz(y_cm, x_cm))

    Both integrals are to be the same

    '''

    if Q_(Iunit0) != Q_('mW/cm3/sr/nm'):  # Q_ makes sure 'mW/sr/cm3/nm' == 'mW/cm3/sr/nm'
        j_nm = conv2(j_nm, Iunit0, 'mW/cm3/sr/nm')

    # Convert ''mW/cm3/sr/nm' to 'mW/cm3/sr/cm_1'
    j_cm = j_nm * 1e7 / wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(j_nm, wavelen)` or
    # `trapz(j_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    j_cm = conv2(j_cm, 'mW/cm3/sr/cm_1', Iunit)

    return j_cm


def convert_emi2nm(j_cm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral emission density in wavenumber base (typically ~mW/cm3/sr/cm_1) to 
    spectral radiance in wavelength base (~mW/cm3/sr/nm)


    Parameters    
    ----------

    j_cm: array
        spectral emission density in Iunit0 unit (~ per wavenumber)

    wavenum: array (cm-1)
        wavenumber

    Iunit0: str
        unit (~ per wavenumber) to convert from 

    Iunit: str
        unit (~ per wavelength) to convert to


    Notes
    -----

    Implementation:

    We use the variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lλ  = - Lν * 1e-7 * ν^2

    '''

    if Q_(Iunit0) != Q_('mW/cm3/sr/cm_1'):
        j_cm = conv2(j_cm, Iunit0, 'mW/cm3/sr/cm_1')

    # Convert 'mW/cm3/sr/cm_1' to 'mW/cm3/sr/nm'
    j_nm = j_cm * 1e-7 * wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    j_nm = conv2(j_nm, 'mW/cm3/sr/nm', Iunit)

    return j_nm


# %% Radiance conversion functions (with changes from ~1/nm in ~1/cm-1)

def convert_rad2cm(l_nm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral radiance in wavelength base (~1/nm) to spectral radiance in
    wavenumber base (~1/cm_1)


    Parameters    
    ----------

    l_nm: array
        spectral radiance in yunit0 unit (~ per wavelength)

    wavenum: array (cm-1)
        wavenumber

    Iunit0: str
        unit (~ per wavelength) to convert from 

    Iunit: str
        unit (~ per wavenumber) to convert to


    Notes
    -----

    Implementation:

    We use the variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lν = - Lλ * 1e7 / ν^2


    Example
    -------

    Validation:

        >>> x_nm, y_nm = s.get('radiance_noslit', 'nm', 'mW/cm2/sr/nm')
        >>> x_cm, y_cm = s.get('radiance_noslit', 'cm', 'mW/cm2/sr/cm_1')
        >>> print(trapz(y_nm, x_nm))
        >>> print(trapz(y_cm, x_cm))

    Both integrals are to be the same

    '''

    if Q_(Iunit0) != Q_('mW/cm2/sr/nm'):  # Q_ makes sure 'mW/sr/cm2/nm' == 'mW/cm2/sr/nm'
        l_nm = conv2(l_nm, Iunit0, 'mW/cm2/sr/nm')

    # Convert ''mW/cm2/sr/nm' to 'mW/cm2/sr/cm_1'
    l_cm = l_nm * 1e7 / wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    l_cm = conv2(l_cm, 'mW/cm2/sr/cm_1', Iunit)

    return l_cm


def convert_rad2nm(l_cm, wavenum, Iunit0, Iunit):
    '''
    Convert spectral radiance in wavenumber base (~1/cm_1) to spectral radiance in
    wavelength base (~1/nm)


    Parameters    
    ----------

    l_cm: array
        spectral radiance in yunit0 unit (~ per wavenumber)

    wavenum: array (cm-1)
        wavenumber

    Iunit0: str
        unit (~ per wavenumber) to convert from 

    Iunit: str
        unit (~ per wavelength) to convert to

    Notes
    -----

    Implementation:

    We use a variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lλ  = - Lν * 1e-7 * ν^2

    '''

    if Q_(Iunit0) != Q_('mW/cm2/sr/cm_1'):  # Q_ makes sure 'mW/sr/cm2/cm_1' == 'mW/cm2/sr/cm_1'
        l_cm = conv2(l_cm, Iunit0, 'mW/cm2/sr/cm_1')

    # Convert 'mW/cm2/sr/cm_1' to 'mW/cm2/sr/nm'
    l_nm = l_cm * 1e-7 * wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    l_nm = conv2(l_nm, 'mW/cm2/sr/nm', Iunit)

    return l_nm


def convert_universal(I, from_unit, to_unit, spec=None,
                      per_nm_is_like='mW/sr/cm2/nm', per_cm_is_like='mW/sr/cm2/cm_1'):
    ''' Return variable var in whatever unit, and converts to to_unit
    Also deal with cases where var is in ~1/nm (per_nm_is_like) or ~1/cm-1
    (per_cm_is_like)


    Parameters    
    ----------

    var: str
        variable to get. Usually 'radiance' or 'radiance_noslit'

    to_unit: str
        unit to convert variable to

    Other Parameters
    ----------------

    spec: :class:`~radis.spectrum.spectrum.Spectrum` object
        needed to get wavenumber in case we need to do a change of variable 
        within the integral

    Notes
    -----

    wavenumber is needed in case we convert from ~1/nm to ~1/cm-1 (requires 
    a change of variable in the integral)

    '''
    Iunit0 = from_unit
    Iunit = to_unit
    try:
        if is_homogeneous(Iunit0, per_nm_is_like) and is_homogeneous(
                Iunit, per_cm_is_like):
            w_cm = spec.get_wavenumber()
            I = convert_rad2cm(I, w_cm, Iunit0, Iunit)
            # note that there may still be a new DimensionalityError
            # raise here if the input was non-sense.
        elif is_homogeneous(Iunit0, per_cm_is_like) and is_homogeneous(
                Iunit, per_nm_is_like):
            w_cm = spec.get_wavenumber()
            I = convert_rad2nm(I, w_cm, Iunit0, Iunit)
            # note that there may still be a new DimensionalityError
            # raise here if the input was non-sense.
        else:  # general case: just convert
            I = conv2(I, Iunit0, Iunit)
    except DimensionalityError:
        raise DimensionalityError(Iunit0, Iunit)

    return I


# %% Test
if __name__ == '__main__':
    from radis.test.test_phys import test_units__fast
    print(('Test :', test_units__fast()))



# -*- coding: utf-8 -*-
"""
Tools to deal with HDF5 cache files
HDF5 cache files are used to cache Energy Database files, and Line Database
files, and yield a much faster access time. 

Routine Listing
---------------


- :func:`~radis.misc.cache_file.check_cache_file`
- :func:`~radis.misc.cache_file.check_not_deprecated`
- :func:`~radis.misc.cache_file.save_to_hdf`

See Also
--------

:func:`~radis.io.hitran.hit2df`, :func:`~radis.io.cdsd.cdsd2df`

-------------------------------------------------------------------------------


"""
# TODO: start using .feather format. Faster than HDF5 for pandas Dataframe, 
# and can be multithreaded.
# see:
# https://gist.github.com/gansanay/4514ec731da1a40d8811a2b3c313f836
# and pd.read_feather(file, nthreads=3)

import os
import h5py
import radis
from warnings import warn
from os.path import exists
from radis.misc.basics import compare_dict, is_float
import pandas as pd


class DeprecatedFileError(DeprecationWarning):
    pass


OLDEST_COMPATIBLE_VERSION = '0.1.22'
'''str: forces to regenerate cache files that were created in a previous version'''

# Just make sure LAST_BACKWARD_COMPATIBLE_VERSION is valid
assert radis.__version__ >= OLDEST_COMPATIBLE_VERSION

# Utils

def get_cache_file(fcache):
    ''' Load HDF5 cache file 
    
    Notes
    -----
    
    we could start using FEATHER format instead. See notes in cache_files.py 
    '''
    
    # Load file
    
    df = pd.read_hdf(fcache, 'df')
    
    # Check file
    
    # ... 'object' columns slow everything down (not fixed format strings!)
    _warn_if_object_columns(df, fcache)
    
    return df

def check_cache_file(use_cached, fcache, verbose):
    ''' Check cache file:

    First test its existence: function of the value of ``use_cached``:

    - if True, ok
    - if 'regen', delete cache file to regenerate it later. 
    - if 'force', raise an error if file doesnt exist. 

    Then look if it is deprecated (we just look at the attributes, the file 
    is never fully read)

    - if deprecated, delete it to regenerate later unless 'force' was used
    '''

    # Test existence of file:
    if not use_cached:
        return                   # we dont want a cache file, no need to test it
    elif use_cached == 'regen':
        if exists(fcache):
            os.remove(fcache)
            if verbose:
                print('Deleted h5 cache file : {0}'.format(fcache))
    elif use_cached == 'force':
        if not exists(fcache):
            raise ValueError('Cache file {0} doesnt exist'.format(fcache))
    else:  # use_cached == True
        pass    # just use the file as is

    # If file is still here, test if it is deprecated:
    # (we just read the attributes, the file is never fully read)
    if exists(fcache):
        if verbose:
            print('Using cache file: {0}'.format(fcache))
        try:
            check_not_deprecated(fcache, metadata={}, current_version=radis.__version__,
                                 last_compatible_version=OLDEST_COMPATIBLE_VERSION)
        except DeprecatedFileError as err:
            if use_cached == 'force':
                raise
            else:   # delete file to regenerate it in the end of the script
                if verbose:
                    print('File {0} deprecated:\n{1}\nDeleting it!'.format(
                        fcache, str(err)))
                os.remove(fcache)

    return


def check_not_deprecated(file, metadata, current_version=None, 
                         last_compatible_version=OLDEST_COMPATIBLE_VERSION):
    ''' Make sure cache file is not deprecated: checks that ``metadata`` is the same,
    and that the version under which the file was generated is valid.

    Parameters
    ----------

    file: str
        a `` .h5`` cache file for Energy Levels 

    metadata: dict
        list of variables used to create the file. If the values dont match,
        an error is raised. 

    current_version: str, or ``None``
        current version number. If the file was generated in a previous version
        a warning is raised. If ``None``, current version is read from 
        :data:`radis.__version__`.

    last_backward_compatible_version: str
        If the file was generated in a non-compatible version, an error is raised.
        Default :data:`~radis.misc.cache_files.OLDEST_COMPATIBLE_VERSION`

    '''

    # Get attributes (metadata+version)
    hf = h5py.File(file, 'r')
    try:
        attrs = dict(hf.attrs)
    except OSError:
        attrs = {}
    finally:
        hf.close()

    # Raise an error if version is not found
    try:
        file_version = attrs.pop('version')
    except KeyError:
        raise DeprecatedFileError('File {0} has been generated in a deprecated '.format(file) +
                                  'version. Delete it to regenerate it on next run')

    # Get current version
    if current_version is None:
        current_version = radis.__version__

    # If file version is anterior to a major change
    # ... Update here versions afterwhich Deprecated cache file is not safe
    # ... (example: a key name was changed)
    if file_version < last_compatible_version:
        raise DeprecatedFileError('File {0} has been generated in a deprecated '.format(file) +
                                  'version ({0}). Oldest compatible version is {1}. '.format(
            file_version, last_compatible_version) +
            'Delete the file to regenerate it on next run')

    # If file version is outdated: Warning, but no error
    if current_version > file_version:
        warn(DeprecationWarning('File {0} has been generated in '.format(file) +
                                'a deprecated version ({0}) compared to current ({1})'.format(
            file_version, current_version) +
            '. Delete it to regenerate it on next run'))
        out = False
    elif current_version == file_version:
        out = True
    else:
        raise ValueError('File generated with a future version?')

    # Compare metadata
    metadata = _h5_compatible(metadata)
    out, compare_string = compare_dict(
        metadata, attrs, verbose=False, return_string=True)
    if out != 1:
        raise DeprecatedFileError('Metadata in file {0} dont match '.format(file) +
                                  'expected values. See comparison below:' +
                                  '\n\tExpected\tFile\n{0}'.format(compare_string))

    return out


def _h5_compatible(a_dict):
    ''' Make dictionary ``a_dict`` h5 compatible '''
    out = {}
    for k, v in a_dict.items():
        if v is None:
            continue   # dont store None
        elif is_float(v):
            out[k] = v
        else:
            out[k] = str(v)   # convert to str
    return out

def _warn_if_object_columns(df, fname):
    '''  'object' columns slow everything down (not fixed format strings!) '''
    objects = [k for k,v in df.dtypes.items() if v == object]
    if len(objects) > 0:
        warn('Dataframe in {0} contains `object` format columns: {1}. '.format(
                fname, objects)+'Operations will be slower. Try to convert them to numeric.')
        

def save_to_hdf(df, fname, metadata, version=None, key='df', overwrite=True):
    ''' Save energy levels to HDF5 file. Add metadata and version 

    Parameters
    ----------

    df: a pandas DataFrame
        data will be stored in the key ``'df'`` 

    fname: str
        ``.h5`` file where to store. 

    metadata: dict
         dictionary of values that were used to generate the DataFrame. Metadata
         will be asked again on file load to ensure it hasnt changed. ``None`` 
         values will not be stored.

    version: str, or ``None``
        file version. If ``None``, the current :data:`radis.__version__` is used. 
        On file loading, a warning will be raised if the current version is 
        posterior, or an error if the file version is set to be uncompatible.
        See :data:`~radis.misc.cache_files.OLDEST_COMPATIBLE_VERSION`

    key: str
        dataset name. Default ``'df'`` 

    overwrite: boolean
        if ``True``, overwrites file. Else, raise an error if it exists.

    Notes
    -----

    ``None`` values are not stored

    '''

    # Check file
    assert fname.endswith('.h5')
    assert 'version' not in metadata
    # ... 'object' columns slow everything down (not fixed format strings!)
    _warn_if_object_columns(df, fname)

    # Update metadata format
    metadata = _h5_compatible(metadata)

    # Overwrite file
    if exists(fname) and not overwrite:
        raise ValueError('File exist: {0}'.format(fname))
    hf = h5py.File(fname, 'w')

    try:

        # Start by adding version
        if version is None:
            version = radis.__version__
        hf.attrs['version'] = version

        # Add metadata
        for k, v in metadata.items():
            hf.attrs[k] = v

    except:
        raise

    finally:
        hf.close()

    # now export dataframe
    df.to_hdf(fname, key, format='fixed', mode='a',
              complevel=1, complib='blosc')
#    df.to_hdf(fname, 'df', format='fixed', mode='a')  # would be 10-20% faster, but take 2x more space


def filter_metadata(arguments, discard_variables=['self', 'verbose']):
    ''' Filter arguments (created with  ``locals()`` at the beginning
    of the script) to extract metadata. Metadata is stored as attributes
    in the cached file:

    - remove variables in ``discard_variables``
    - remove variables that start with ``'_'``
    - remove varibles whose value is ``None``

    Parameters
    ----------

    arguments: dict
        list of local variables. For instance::

            arguments = locals()

    discard_variables: list of str
        variable names to discard

    Returns
    -------

    metadata: dict
        a (new) dictionary built from arguments by removing ``discard_variables``
        and variables starting with ``'_'``

    Examples
    --------

    How to get only function argument::

        def some_function(*args):
            metadata = locals()     # stores only function arguments because it's the first line

            ...

            metadata = filter_metadata(metadata)
            save_to_hdf(df, fname, metadata=metadata)

            ...

    '''

    metadata = {k: v for (k, v) in arguments.items() if not k.startswith('_')}
    metadata = {k: v for (k, v) in metadata.items()
                if k not in discard_variables}
    metadata = {k: v for (k, v) in metadata.items() if v is not None}

    return metadata



''' Install file for RADIS

Typical install procedure, plus:
    
- auto-convert README.rst to long_description, removing some sphinx-only syntax 
so it can be rendered by PyPi
- read version number from __version__.txt
- some dependencies should be installed manually (typically: all packages with 
compiled components such as numpy, pandas, etc.)


Examples
--------

Install (normal, use-only)::

    python setup.py install
    
Or (create an alias, so you can still edit)::
    
    python setup.py develop

Notes
-----

For developers:

when creating a new version, just update the __version__.txt file

to register it on Pypi see register.py::
    
    python register.py 
 

'''
from __future__ import print_function
from __future__ import absolute_import
from setuptools import setup, find_packages
import codecs
import io
import re
import os
from os.path import dirname, join
import sys
from setuptips import yield_sphinx_only_markup

# Check mandatory modules that we want users to install themselves
try:
    import numpy 
    import scipy
    import matplotlib
    import pandas
except ImportError:
    raise ImportError('Please install these librairies first (with Anaconda is '+\
                      'strongly recommended) \n >>> conda install numpy scipy '+\
                      'matplotlib pandas')

# Build description from README (PyPi compatible)
description = 'A non-equilibrium Radiative Solver for HITRAN-like database species '
readme_lines = codecs.open('README.rst', encoding="utf-8").readlines()
long_description = ''.join(yield_sphinx_only_markup(readme_lines))

# Read version number from file
with open(join(dirname(__file__),'radis', '__version__.txt')) as version_file:
    __version__ = version_file.read().strip()
    
# Main install routine
setup(name='radis',
      version=__version__,
      description=description,
    	long_description=long_description,
      url='https://github.com/radis/radis',
      author='Erwan Pannier',
      author_email='erwan.pannier@gmail.com',
      license='GNU Lesser General Public License v3 (LGPLv3)',
      keywords=["spectrum", "infrared", "spectra", "radiation", "nonequilibrium"],
      packages=find_packages(),
      install_requires=[
				 'mpldatacursor',
#                        'numpy',          # let the user install it 
#                        'scipy',         # let the user install it  
#                        'matplotlib',    # let the user install it 
#                        'pandas',        # let the user install it 
#                        'sympy',         # let the user install it 
                        'pint>=0.7.2',  # Unit aware calculations
                        'publib>=0.2.1', # Plotting styles for Matplotlib
                        'plotly>=2.5.1',    # for line survey HTML output  
                        # TODO: update to >=2.5.1 as soon as https://github.com/plotly/plotly.py/issues/963 is fixed
                        'termcolor',     # terminal colors
                        'six',  # python 2-3 compatibility
                        'configparser', 
                        'astroquery',   # to fetch HITRAN databases
                        'json-tricks',   # to deal with non jsonable formats
                        'numpydoc',     # for Jedi (autocompletion) to recognize
                                        # numpy docstrings
                        'tables',       # for pandas to HDF5 export
                        'pytest',       # to run test suite
                        'h5py',         # to write HDF5 files
				],
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent"],
	  include_package_data=True,  # add non .py data files in MANIFEST.in
      package_data={'radis': ['radis/phys/units.txt']},
      zip_safe=False,  # impossible as long as we have external files read with __file__ syntax
      platforms='any')


# -*- coding: utf-8 -*-
"""

Physical constants

Note : more units available in :
- scipy.constants
- sympy.physics.units


-------------------------------------------------------------------------------


"""

# %% SI units

eV = 1.60217657e-19  # J
'''float: electron volt (J)'''

h = 6.62606957e-34
'''float: Planck constant (m2.kg.s-1 = J.s)'''

k_b = 1.3806488e-23 
'''float: Boltzmann constant (m2.kg.s-2.K-1)'''

c = 2.99792458e8
'''float: light velocity (m/s)'''

a_0 = 5.29177e-11
'''float: First Bohr radius (m)'''

m_e = 9.10938291e-31
'''float: Electron mass (kg)'''

Ry = 13.60569253
'''float: Rydberg unit of energy (eV)'''

Av = 6.02214129e23
'''float: Avogadro constant (molecules / mol)'''
Na = Av  # just an alias

eps_0 = 8.854187817e-12
'''float: Vacuum permittivity (Farads / m)'''

hc_k = h*c/k_b*100  # 
'''float:  cm to K conversion (~ 1.44 cm.K), used in Boltman factors'''


# %% HITRAN (CGS) units

k_b_CGS = 1.380648813E-16
'''float: Boltzman constant (erg / K)'''

c_CGS = 2.99792458e10
'''float: Light velocity (cm / s)'''

h_CGS = 6.626196e-27
'''float: Planck constant (erg * s)'''
